////////////////////////////////////////////////////////////////////////
/*! 
 *  \file   Pack.h
 *  \brief  Templates to pack/unpack channel bits into bytes from the tiff library.
 *  \author JD Gascuel
 * 
 * Copyright JD Gascuel, 2006.
 * 
 * Jean-Dominique.Gascuel <at> imag.fr
 *
 * This software is a computer program whose purpose is to add TIFF image
 * read/write capabilities to the Qt library, hence to any Qt application.
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 * TiffIO is based on the libtiff 3.8.0, with the following copyright :
 *      Copyright (c) 1988-1997 Sam Leffler
 *      Copyright (c) 1991-1997 Silicon Graphics, Inc.
 * See the libtiff-3.78.0/ subdirectory.
 */
////////////////////////////////////////////////////////////////////////
// File History:
//
// $Id: Pack.h 13 2010-06-25 16:31:07Z gascuel $
//
// 2006-04-03 : Creation.

#ifndef PACK_H
#define PACK_H

// Auto-detect needs for swaping byte...
// We just undo what libtiff mixed-up when bps==16, and the native ordering of
// the file does not match the one of the processor...
static union { short word; char byte[2]; } needSwabTest = { 0x7F00 };
#define NEED_SWAB_KLUDGE bool(needSwabTest.byte[1])

//////////////////////////////////////////////////////////////////////////////
/*! \brief Decompress packed bits into words.
 *
 * Extract \p bps bits packet from the \p compressed buffer ib tiff \p format, 
 * extend them to \p Word, and write them to the \p buffer.
 *
 * If \p format is:
 *  + \p SAMPLEFORMAT_IEEEFP : 32 bits word are interpreted as float, and converted to int.
 *  + \p SAMPLEFORMAT_INT    : a sign extension is performed.
 *  + \p SAMPLEFORMAT_UINT   : nothing special happends.
 *
 * \b Note:
 *    No decompression in place is possible, because the result is larger than
 *    the original data. You should provide differents buffers.
 */
template<class Word, bool signedResult, bool high>
void unpack(uint bps, uint16 format, int width, const unsigned char * compressed, Word* buffer)
{
    Q_ASSERT(sizeof(ulong) >= sizeof(Word));
    Q_ASSERT(compressed && buffer && (void*)compressed != (void*)buffer);

    ulong mask = (bps<32) ? (1<<bps)-1 : 0xffffffff;
    ulong sign = 1<<(bps-1);
    bool swab = NEED_SWAB_KLUDGE && (bps == 16);

    ulong c=0;                           // Current carry (already read bits)
    uint  b=0;                           // Current bits # in carry.
    for(int i=0; i<width; ++i)
    {
        // Re-assemble -------------------------------------------------------
        while( b<bps ) // Not enough bits ? read more
        {
            c = ulong(*compressed++)
              | (c<<8)
              ;
            b += 8;                 // 8 bits read in.
        }
        b -= bps;                   // Update current length.
        uint s = mask & (c>>b);

        if( swab )
            s = (s>>8)&0xFF | ((s&0xFF)<<8);

        // Do ourself sign extension, if needed :
        if( format == SAMPLEFORMAT_INT)
        {
            if( s >= sign )       // eg. bps=8: if >= 128
                s -= (mask+1);                      // eg. bps=8: 240 --> -16
            if( !signedResult )
                s += 1<<(8*sizeof(Word)-1);
        }

        // Transform floats to 8 bits int
        if( format==SAMPLEFORMAT_IEEEFP && bps == 32)
        {
            float v = ((float*)(&s))[0];    // No chance for this to be portable... Suggestions ?
            s = Word(0.5 * mask * v);       // bps=8: +127 .. -127 (or -128 ?)

            if( !signedResult )             // back to 0..255, if needed.
                s += 1<<(8*sizeof(Word)-1);
        }

        // Chop bps-bits sample to target type size --------------------------
        if( high )
        {
            if( bps > 8*sizeof(Word) )
                s >>= (bps-8*sizeof(Word));
            else if( bps < 8*sizeof(Word) )
                s <<= (8*sizeof(Word)-bps);
        }

        buffer[i] = (Word)s;
    }
}

//////////////////////////////////////////////////////////////////////////////
/*! \brief Compress words by packing lower bits into bytes.
 *
 * Pack \p bps bit samples taken from \p width \p Word, into a buffer of
 * bytes.
 * Last byte is zero-filled. If you need to have line alignement, call
 * this pack function for each line.
 *
 * \b Note:
 *    Because packed buffer is smaller than the unpacked one, we could have
 *    compression in place (ie. \p buffer == \p compressed).
 * \b Note:
 *    Taking lower bits (\p high == false) is nice for color indexes.
 *    Taking higher bits (\p high == true) is nice for componants.
 */
template<class Word, bool high>
void pack(uint bps, int width, const Word * buffer, unsigned char* compressed)
{
    Q_ASSERT(compressed && buffer && (void*)compressed != (void*)buffer);

    ulong mask  = (bps<32) ? (1<<bps)-1 : 0xffffffff;
    int   shift = 8*sizeof(Word) - bps;
    Q_ASSERT(shift >= 0);

    ulong c = 0;            // Carry (bits read so forth).
    int   b = 0;            // # of bits in carry.
    int   o = 0;            // # of bytes written so forth.
    while(width > 0)
    {
        while( b<8 )        // Not a full byte to write in carry ?
        {
            if( high )
            {
                // we keep the most significativ bits. Ie. to pack to 4 bits,
                // we should have values in 0..255, divided by 16 on save.
                c = (((*buffer++)>>shift) & mask)   // Read in bps higher bits
                  | (c<<bps)                        // and accumulate with previous bits.
                  ;
            }
            else
            {
                // we keep the less significativ bits. Ie. to pack to 4 bits,
                // we should have values in 0..15.
                c = ((*buffer++) & mask)        // Read in bps lower bits
                  | (c<<bps)                    // and accumulate with previous bits.
                  ;
            }
            b+=bps;                         // bookeeping.
            width--;                        // We did read a sample.
        }

        // Output left-most bits (those read first) :
        b -= 8;
        compressed[o++] = (unsigned char)(c>>b);
    }
    
    // Write trailer byte(s), if any.
    while( b > 8 )      // Several trailer bytes ? write as usual.
    {
        b -= 8;
        compressed[o++] = (unsigned char)(c>>b);
    }
    if( b > 0 )         // Uncomplere byte ? add 0-bit trailer
    {
        compressed[o++] = (unsigned char)(c << (8-b));
    }

    if( NEED_SWAB_KLUDGE && (bps == 16) )
    {
        unsigned short* pw = (unsigned short*)compressed;
        for(int i=0; i<o; i+=2)
        {
            unsigned short w = pw[0];
            pw[0] = ((w & 0xFF) << 8) | ((w>>8) & 0xFF);
        }
    }
}
#endif // PACK_H
