////////////////////////////////////////////////////////////////////////
/*! 
 *  \file   TiffIO.cpp
 *  \brief  Plugin to enable TIFF I/O in Qt 3.x and 4.x.
 *  \author JD Gascuel
 * 
 * Copyright JD Gascuel, 2005.
 * 
 * Jean-Dominique.Gascuel@imag.fr
 *
 * This software is a computer program whose purpose is to add TIFF image
 * read/write capabilities to the Qt library, hence to any Qt application.
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 * TiffIO is based on the libtiff 3.7.2, with the following copyright :
 *      Copyright (c) 1988-1997 Sam Leffler
 *      Copyright (c) 1991-1997 Silicon Graphics, Inc.
 * See the libtiff-3.7.2/ subdirectory.
 */
////////////////////////////////////////////////////////////////////////
// File History:
//
// $Id: TiffIO.cpp 13 2010-06-25 16:31:07Z gascuel $
//
// 2004-11-03 : Creation.
//              Support reading of RGBA and RGB bits true color images
//              with up to 16 bits per component (automatically downscaled 
//              to 3x8 bits colors), up to 16 bits gray images (in MINISWHITE 
//              or MINISBLACK format, downscaled to 8-bit grays), 
//              Color palette with up to 8 bits.
//              Support all Paletted/RGB/Gray TIFF encoding scheme (in 
//              particular LZW and ZIP), 
//              except JPEG compression, and LAB or CMYK color spaces.
//
//              Support writting of Qt images with 8-bpp, and 24 and 32-bpp.
//              Always use LZW compression. Automatically switches to MINISBLACK
//              when saving 8-bit all gray image. Use PALETTE mode for 8-bit images,
//              and RGBA mode for 24/32 bit images.
//
// 2004-11-04 : Simplified, and added support for 8-bit LAB and CYMK images.
//              Added support for non-file IO, so .gz images is working.
//              Added support for fileBigEndians as well as littleEndians file.
//
// 2004-11-22 : FIX : regexp should start with a ^, or wrong image could
//              be declared to be TIFF images.
//
// 2004-05-27 : FIX from Till Oliver Knoll, when calling QtTIFFOpen on a memory
//              stream.
// 2005-01-31 : FIX more compatibility issues on HP-UX and AIX, 
//              thanks to Till Oliver Knoll.
//
// 2005-09-24 : Initial port to Qt 4.0.1
//
// 2006-01-06 : FIX bit order bug in bitmaps (this different from qtBigEndian).
//
// 2006-03-27 : FIX setAlphaBuffer for Qt3 (thanks to Till Oliver Knoll).
//              FIX raw vs decompressed 11bits pixels with PixarLog compressor.

// Beware: some version of libtiff forgets about the extern "C" for some
// fucntion. So make sure we do it :
extern "C" {
#   include "tiffio.h"
}

#include <string.h>
#include <qglobal.h>
#include <stdlib.h>

#include "Pack.h"

#if QT_VERSION < 0x040000
#   include <qimage.h>
#   include <qiodevice.h>
#   define HANDLER_CLASS QImageIO
#   define HANDLER_DEVICE ioDevice
#else
#   include <QString>
#   include <QImage>
#   include <QImageIOHandler>
#   include <QVariant>
#   include <QRegExp>
#   define HANDLER_CLASS QImageIOHandler
#   define HANDLER_DEVICE device
#   define Q_UINT32 quint32
#   define latin1   toLatin1
#endif

#define GETDEVICE                                               \
    HANDLER_CLASS *io = reinterpret_cast<HANDLER_CLASS*>(fd);   \
    Q_ASSERT(io);                                               \
    QIODevice* dev = io->HANDLER_DEVICE();                      \
    Q_ASSERT(dev);

////////////////////////////////////////////////////////////////////////
/*
 * TIFF Library Qt-specific Routines.  Adapted from tif_win32.c, 
 * Sam Leffler/SGI 1997, and tif_unix 4/5/95 by
 * Scott Wagner (wagner@itek.com), Itek Graphix, Rochester, NY USA
 */

static tsize_t
QtReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
    GETDEVICE
    if( dev->isReadable() )
#if QT_VERSION < 0x040000
        return (tsize_t) dev->readBlock((char *)buf, size);
#else
        return (tsize_t) dev->read((char *)buf, size);
#endif
    else
        return 0;
}

static tsize_t
QtWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
    GETDEVICE
    if( dev->isWritable() )
#if QT_VERSION < 0x040000
        return (tsize_t) dev->writeBlock((char *)buf, size);
#else
        return (tsize_t) dev->write((char *)buf, size);
#endif
    else
        return 0;
}

static toff_t
QtSizeProc(thandle_t fd)
{
    GETDEVICE

    // Beware : there is a bug in QT, size() re-read what
    // disk says, and do not take into account any buffered
    // write pending...
    //   (QFile uses fwrite, that do buffering over FILE.
    //    But the fh is a private member we do not have access to).
#if QT_VERSION < 0x040000
    if( dev->isWritable() )
        dev->flush();
#else
    // Stupid folks from Trolltech suppressed QIODevice::flush()
    // from Qt4...
    // So we have to find a kludge to work around that:
    if( dev->isWritable() && ! (dev->openMode() & QIODevice::Unbuffered) )
    {
        dev->close();
        // Re-open in unbuffered mode. Do it in ReadWrite, so data
        // already written is not truncated. And not in Appen mode
        // too, that will clobber any further seek...
        dev->open( QIODevice::ReadWrite | QIODevice::Unbuffered );
    }
#endif
    
    toff_t pos  = (toff_t) dev->size();
    return pos;
}

static toff_t
QtSeekProc(thandle_t fd, toff_t offset, int where)
{
    GETDEVICE

    switch(where) {
	case SEEK_CUR:
#if QT_VERSION < 0x040000
        offset += dev->at();
#else
        offset += dev->pos();
#endif
		break;
	case SEEK_END:
        offset = QtSizeProc(fd) - offset;
		break;
    }

    if(
#if QT_VERSION < 0x040000
        dev->at(offset)
#else
        dev->seek(offset)
#endif
        )
        return offset;
    else
    	return 0;
}

/*
 * Closing the file just needs to get the right pointer, and close the right QIODevice.
 */
static int
QtCloseProc(thandle_t fd)
{
    GETDEVICE
    dev->close();
	return 0;
}

static int
QtMapProc(thandle_t, void**, toff_t*)
{
    qDebug("TIFF map proc not implemented");
    return 0;   // Always failed.
}

static void
QtUnMapProc(thandle_t, void*, toff_t)
{
    qWarning("TIFF unmap proc not implemented");
}

/*
 * Open a TIFF file descriptor for read/writing.
 * Note that TIFFFdOpen and TIFFOpen recognise the character 'u' in the mode
 * string, which forces the file to be opened unmapped.
 */
TIFF*
QtTIFFOpen(HANDLER_CLASS *io, char *mode)
{
    // From Qt 4.x, there is no way we can retrieve the fileName of a given QIODevice.
    // (which is meaningfull for array buffer, or sockets, but a pity for good old
    //  disk files).
    static char *fileName = "QImage";
	TIFF* tif = TIFFClientOpen(fileName, mode, (thandle_t)io,
			QtReadProc, QtWriteProc,
			QtSeekProc, QtCloseProc, QtSizeProc,
			QtMapProc,  QtUnMapProc);
	return tif;
}

////////////////////////////////////////////////////////////////////////

static void QtTIFFErrorHandler(const char* module, const char* fmt, va_list ap)
{
    char buffer[1024];
#ifdef WIN32
    // Windows people are stupid. As usual:
    _vsnprintf(buffer, sizeof buffer, fmt, ap);
#else
    vsnprintf(buffer, sizeof buffer, fmt, ap);
#endif

    // TIFF codec should never crash the main App. Just generate warnings !
    qWarning(QString("%1: FATAL error : %2")
        .arg(module)
        .arg(buffer)

        .latin1()
    );
}

static void QtTIFFWarningHandler(const char* module, const char* fmt, va_list ap)
{
    char buffer[1024];
#ifdef WIN32
    // Windows people are stupid. As usual:
    _vsnprintf(buffer, sizeof buffer, fmt, ap);
#else
    vsnprintf(buffer, sizeof buffer, fmt, ap);
#endif

    qWarning(QString("%1: Warning : %2")
        .arg(module)
        .arg(buffer)
        .latin1()
    );
}

////////////////////////////////////////////////////////////////////////
// TIFF specific functions :
//    Handle CIE L,a,b color coordinates to RGB conversion.

// This is from tif_getimage.c (line 54), but sometimes failes to link
// for unknown reason... So add a copy here...
static TIFFDisplay display_sRGB = {
	{			/* XYZ -> luminance matrix */
		{  3.2410F, -1.5374F, -0.4986F },
		{  -0.9692F, 1.8760F, 0.0416F },
		{  0.0556F, -0.2040F, 1.0570F }
	},	
	100.0F, 100.0F, 100.0F,	/* Light o/p for reference white */
	255, 255, 255,		    /* Pixel values for ref. white */
	1.0F, 1.0F, 1.0F,	    /* Residual light o/p for black pixel */
	2.4F, 2.4F, 2.4F,	    /* Gamma values for the three guns */
};
static float   refWhite[3];
static TIFFCIELabToRGB* cielab = 0;
 
void initCIELabConversion(TIFF* tif)
{
	float   *whitePoint;

	if( !cielab)
    {
		cielab = (TIFFCIELabToRGB *)_TIFFmalloc(sizeof(TIFFCIELabToRGB));
        Q_CHECK_PTR(cielab);
    }

	TIFFGetFieldDefaulted(tif, TIFFTAG_WHITEPOINT, &whitePoint);
	refWhite[1] = 100.0F;
	refWhite[0] = whitePoint[0] / whitePoint[1] * refWhite[1];
	refWhite[2] = (1.0F - whitePoint[0] - whitePoint[1])
		      / whitePoint[1] * refWhite[1];
	if (TIFFCIELabToRGBInit(cielab, &display_sRGB, refWhite) < 0)
        qWarning("Failed to initialize CIE L*a*b*->RGB conversion state.");
}

////////////////////////////////////////////////////////////////////////

#define TIFF2QT(c)  qRgba(TIFFGetR(c), TIFFGetG(c), TIFFGetB(c), TIFFGetA(c))

void commonReadTIFF(HANDLER_CLASS* handler, QImage& img, int& status, QString& description)
{
    TIFFSetErrorHandler(QtTIFFErrorHandler);
    TIFFSetWarningHandler(0);                           // Strip warnings while readind TIFF's dir.
        TIFF *tif = QtTIFFOpen(handler, "rBm");         // LZW bogus in H or L mode...
    TIFFSetWarningHandler(QtTIFFWarningHandler);

    if( tif )
    {
        
        // Any special cases for some de-compressors that have special behaviour ?
        // Like in GetRGBAImage, ask politly decompressor to be as near as possible to
        // standard cases...
        {
            uint16 compression;
            TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression);

            // Fake behaviour of TIFFRGBAImageBegin().
            if( compression == COMPRESSION_SGILOG || compression == COMPRESSION_SGILOG24 )
            	TIFFSetField(tif, TIFFTAG_SGILOGDATAFMT, SGILOGDATAFMT_8BIT);

            // Pixar codec also defaults to output raw encoded 11bits data.
            if( compression == COMPRESSION_PIXARLOG || compression == COMPRESSION_PIXARFILM )
                TIFFSetField(tif, TIFFTAG_PIXARLOGDATAFMT, PIXARLOGDATAFMT_8BIT);
        }

        // TIFFIsBigEndian() is a new function, known from libtiff 3.8.0
        // (or was it 3.7.0 ???). I need to be backward compatile with tiff
        // 3.6.1 (don't ask). So I have to re-implement that. And because
        // the TIFF struct is private, I have to re-read the file header,
        // where I still know what the device handler is. That is right here !
        uint16 tif_magic;

        // Mandatory fields :
        uint32 width=0, height=0;
        int success = QtSeekProc((thandle_t)handler, 0, SEEK_SET)==0
                   && QtReadProc((thandle_t)handler, &tif_magic, sizeof tif_magic)==2
                   && TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,      &width)
                   && TIFFGetField(tif, TIFFTAG_IMAGELENGTH,     &height)
                   ;

        // Fields for which we want to try some not-too-stupide default values :
        uint16 bps       = 1,
               spp       = 1, 
               mode      = PHOTOMETRIC_MINISBLACK,
               format    = SAMPLEFORMAT_UINT,
               planar    = PLANARCONFIG_CONTIG,
			   preMultiplied = 0;
        TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   &bps);
        TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
        TIFFGetField(tif, TIFFTAG_PHOTOMETRIC,     &mode);
        TIFFGetField(tif, TIFFTAG_SAMPLEFORMAT,    &format);
        TIFFGetField(tif, TIFFTAG_PLANARCONFIG,    &planar);
		TIFFGetField(tif, TIFFTAG_MATTEING,        &preMultiplied);
		

        //--------------------------------------------------------------------
        // Do we need byte swapping ? (also done in libtiff, and in Pack.h 8-()
        // Ask Qt about qtBigEndian and word size of the platform :
#if QT_VERSION < 0x040000
        bool qtBigEndian;
        int wordSize;
        qSysInfo( &wordSize, &qtBigEndian );
#else
        bool qtBigEndian = QSysInfo::BigEndian;
#endif
        bool fileBigEndian = tif_magic == TIFF_BIGENDIAN;
        bool badIndians = fileBigEndian != qtBigEndian;

        //--------------------------------------------------------------------
        // The only supported mode for transparency is RGBA.
        // Other stuff (as gray+alpha, or palatted+alpha, or CIELab+alpha, etc.)
        // are not supported...
        bool hasAlpha = ((mode==PHOTOMETRIC_RGB) && (spp==4));

        //--------------------------------------------------------------------
        // Fill-up a nice humane readable description of all this mess,
        // for debugging purpose, and to pass to the calling image.
        static const char *photoNames[] = {
            "min-is-white",				/* PHOTOMETRIC_MINISWHITE */
            "min-is-black",				/* PHOTOMETRIC_MINISBLACK */
            "RGB",				        /* PHOTOMETRIC_RGB */
            "paletted",	                /* PHOTOMETRIC_PALETTE */
            "transparency mask",		/* PHOTOMETRIC_MASK */
            "separated",				/* PHOTOMETRIC_SEPARATED */
            "YCbCr",					/* PHOTOMETRIC_YCBCR */
            "?",
            "CIE L*a*b*",				/* PHOTOMETRIC_CIELAB */
        };
#define	NPHOTONAMES	(sizeof (photoNames) / sizeof (photoNames[0]))
        const char *modeString = (mode<NPHOTONAMES) ? photoNames[mode] : "?";
        const char *signedString = (format == SAMPLEFORMAT_UINT ) ? "unsigned"
                                    : (format == SAMPLEFORMAT_INT )  ? "signed"
                                    : (format == SAMPLEFORMAT_IEEEFP ) ? "float"
                                    : "<unknown>"
                                    ;

        description = QString("%1x%2, %5 x %3-bit %4 %9pixels, %6 color mode%7%8")
            .arg(width).arg(height)
            .arg(bps).arg(signedString).arg(spp)
            .arg(modeString)
            .arg(hasAlpha  ? preMultiplied ? ", with pre-multiplied alpha" : ", with alpha" : "")
            .arg( ((planar != PLANARCONFIG_CONTIG) && (spp>1)) ? ", organised in separated planes" : "")
            .arg( (bps>8) ? fileBigEndian ? "bigEndian " : "littleEndian " : "" )
            ;

        if( (spp==3) || (spp==4) )
        {
#if QT_VERSION < 0x040000
            img.create(width, height, 32);
            img.setAlphaBuffer( hasAlpha );
#else
			QImage::Format imgFormat 
                = hasAlpha ? preMultiplied ? QImage::Format_ARGB32_Premultiplied
                                           : QImage::Format_ARGB32
                           : QImage::Format_RGB32;
            img = QImage(width, height, imgFormat);
            // Note: in Qt4, Format_ARGB32 is the mandatory way to create alpha channeled images.
#endif

            if( img.isNull() )
            {
                qWarning("Cannot allocate 32 bit image %dx%d", (int)width, (int)height);
                return;
            }
 
            // Helpfull function, that handle any LAB/CYMK color conversion as well: stick with it !
            // (But only when in 8 bps and without alpha ... too bad!)
            if( bps == 8 || (bps==16 && mode==PHOTOMETRIC_RGB && !hasAlpha) )
            {
                success &= TIFFReadRGBAImageOriented(tif, width, height, (uint32*)img.bits(), ORIENTATION_TOPLEFT);
                if( success && (TIFF2QT(0x11223344) != 0x11223344) )
                    for(uint i=0; i<width; i++)
                        for(uint j=0; j<height; ++j)
                            img.setPixel(i,j, TIFF2QT( img.pixel(i,j) ));
            }
            else
            {
                // When there is 16bits by pixels, TIFFReadRGBAImageOriented() knows only about
                // RGB mode. When there is more, it don't know anything.
                // So, we just do it ourself, rounding all values
                // to 8bits *before* color conversion. This is (very slightly) bad, but
                // enables us to re-use libtiff conversion functions. Which is *very* good.
                //
                // Possible modes for multi-sample-per-pixel images are :
                //  + PHOTOMETRIC_RGB    : done already up to 16 bits int. Neither 32 or floats.
                //  + PHOTOMETRIC_CIELAB : tif_color.c providdes conversion.
                //  + PHOTOMETRIC_YCBCR  : also. But there is complicated subsampling here. -> NYI.
                //  + PHOTOMETRIC_SEPARATED : Inks, like CMJK formats... --> NYI.
                //  + PHOTOMETRIC_LOGL and PHOTOMETRIC_LOGLUV :
                //          Use SGILOGDATAFMT_8BIT to ask tif_luv.c to convert to RGB 8 bits ???

                // Check we do know how to decompress and down size sample to 8 bits unsigned values :
                if( format != SAMPLEFORMAT_UINT && format != SAMPLEFORMAT_INT && format != SAMPLEFORMAT_IEEEFP )
                    goto Failed;
                if( format == SAMPLEFORMAT_IEEEFP && bps != 32 )
                    goto Failed;
                if( planar != PLANARCONFIG_CONTIG )
                    goto Failed;

                // Any color conversion init ?
                if( mode == PHOTOMETRIC_CIELAB )
                    initCIELabConversion(tif);

                // The allocate tmp space (just one line, so won't consume a lot).
                unsigned char* scanlineBuffer = new unsigned char[TIFFScanlineSize(tif)];
                Q_CHECK_PTR(scanlineBuffer);
                unsigned char* decompressed = new unsigned char[width*spp];
                Q_CHECK_PTR(decompressed);

                for(uint y=0; y<height; ++y)
                {

                    success &= ( TIFFReadScanline(tif, scanlineBuffer, y, 0) > 0);
                    if( !success ) break;

                    // On-the-fly packed bit decompression and truncation of 8bits values
                    unpack<unsigned char,false,true>(bps, format, spp*width, scanlineBuffer, decompressed);

                    QRgb*   q = (QRgb*)img.scanLine(y);
                    switch(mode)
                    {
                    case PHOTOMETRIC_RGB:
                        {
                            unsigned char* p = (unsigned char*)decompressed;

                            // Just copy pixels, with or without alpha.
                            if( spp == 4 )
                                for(uint x=0; x<width; ++x, p+=spp)
                                    q[x] = qRgba(p[0], p[1], p[2], p[3]);
                            else
                                for(uint x=0; x<width; ++x, p+=spp)
                                    q[x] = qRgb(p[0], p[1], p[2]);
                        }
                        break;

                    case PHOTOMETRIC_CIELAB:
                        {
                            char* p = (char*)decompressed; // a and b are signed, L is not...
                            for(uint x=0; x<width; ++x, p+=spp)
                            {
                                float  X=0, Y=0, Z=0; TIFFCIELabToXYZ(cielab,
                                        ((unsigned char* )p)[0], // *NO* sign extension for L !
                                        p[1], p[2], 
                                        &X, &Y, &Z);
                                uint32 r=0, g=0, b=0; TIFFXYZToRGB(cielab, X, Y, Z, &r, &g, &b);
                                q[x] = qRgb(r,g,b);
                            }
                        }
                        break;

                    default:
                        goto Failed;
                    }
                }

                delete[] decompressed;
                delete[] scanlineBuffer;
                if(cielab) { _TIFFfree(cielab); cielab=0; }
            }
        }
        else if( spp==1 )
        {
            // Qt does not support more than 256 colors.
            int colors = (bps <8) ? 1<<bps : 256;

            // Beware of special parameters for bit bitmaps:
            // NOTE : Whatever the fillOrder is in the TIFF header,
            //        libtiff will automatically reverse it anyway.
#if QT_VERSION < 0x040000
            img.create(width, height, 
                       (bps>1) ? 8 : 1, 
                       colors, 
                       (bps>1) ? QImage::IgnoreEndian
                               : QImage::BigEndian);
#else
            img = QImage(width, height,
                (bps>1) ? QImage::Format_Indexed8 
                        : QImage::Format_Mono);
#endif
            if( img.isNull() )
            {
                qWarning("Cannot allocate %d-bit image %dx%d", bps, (int)width, (int)height);
                return;
            }

            // Reading and translating color map, if any ?
            if( mode==PHOTOMETRIC_PALETTE )
            {
                uint16 *red=0, *green=0, *blue=0;
                if( ! TIFFGetField(tif, TIFFTAG_COLORMAP, &red, &green, &blue) )
                    qWarning("Missing palette data in color-palette TIFF image");
                else
                {
                    Q_CHECK_PTR( red ); Q_CHECK_PTR( green ); Q_CHECK_PTR( blue);
                    if( badIndians )
                    {
                        TIFFSwabArrayOfShort(red,   colors);
                        TIFFSwabArrayOfShort(green, colors);
                        TIFFSwabArrayOfShort(blue,  colors);
                    }
                    img.setNumColors(colors);
                    for(int c=0; c<colors; ++c)
                        img.setColor(c, qRgb(red[c]>>8,green[c]>>8,blue[c]>>8));
                }
            }
            else if( (mode==PHOTOMETRIC_MINISWHITE) || (mode==PHOTOMETRIC_MINISBLACK) )
            {
                float s = 255.0f/(colors-1);
                img.setNumColors(colors);
                for(int c=0; c<colors; ++c)
                    img.setColor(
                        (mode==PHOTOMETRIC_MINISBLACK) ? c : (colors-1-c),
                        qRgb((int)(s*c + 0.5f), (int)(s*c + 0.5f), (int)(s*c + 0.5f)));
            }
            else
                qWarning("Invalid palette in single channel TIFF image (mode=%d)", mode);


            // Reading indexes line by line. Unpack any strange format to 8bits (Qt is limited to 256 colors).
            if( bps == 8 || bps == 1)                          // Both 1-bit and 8-bits don't need extra works.
            {
                for(uint y=0; success && (y<height); ++y)
                    success &= ( TIFFReadScanline(tif, img.scanLine(y), y) > 0);
            }
            else
            {
                unsigned char* scanlineBuffer = new unsigned char[TIFFScanlineSize(tif)];
                unsigned char* decompressed = new unsigned char[width];
                Q_CHECK_PTR(decompressed);

                for(uint y=0; y<height; ++y)
                {
                    success &= ( TIFFReadScanline(tif, scanlineBuffer, y) > 0);
                    if( !success ) break;

                    // All modes do use a palette...
                    // *BUT* when we have less than 8-bits, it is color index, when we have more,
                    // we should round to have 0.255 values.
                    if( bps < 8 )
                        unpack<unsigned char,false,false>(bps, format, width, scanlineBuffer, decompressed);
                    else
                        unpack<unsigned char,false,true >(bps, format, width, scanlineBuffer, decompressed);

                    for(uint x=0; x<width; ++x)
                        img.setPixel(x, y, decompressed[x]);
                }
                delete[] decompressed;
            }
        }
        else
        {
Failed:
            qWarning( QString("Unsupported pixel format while reading TIFF %1")
                .arg(description)
                .latin1());

            success = 0;
            img = QImage();
        }

        TIFFClose(tif);

        status = success ? 0 : -2;      // Everything went OK ?
    }
    else
    {
        img = QImage();
        status = -1;                    // Error reading image...
    }
}

void commonWriteTIFF(HANDLER_CLASS* handler, const QImage& img, int& status)
{
    TIFFSetErrorHandler(QtTIFFErrorHandler);
    TIFFSetWarningHandler(QtTIFFWarningHandler);

    TIFF *tif = QtTIFFOpen(handler, "wBm");   // LZW bogus in H or L mode...

    if( tif )
    {
        //---------------------------------------------------------------------
        // Retrieve Qt3's and Qt4's parameters, and translate that to TIFF dialect.

        uint16 bps = 8;
        int quality = 0xffff;   // (ushort)-1 is supposed to be the default value
        const char* params = 0; // No special parameter string by default.
        
#if QT_VERSION < 0x040000
        quality = handler->quality();
        params = handler->parameters();
#else
        QByteArray description;    // Should not be deleted until the end of this function !
        if( handler->supportsOption( QImageIOHandler::Quality ) ) 
			quality = handler->option( QImageIOHandler::Quality ).toInt();
        if( handler->supportsOption( QImageIOHandler::Description ) )
        {
            description = handler->option( QImageIOHandler::Description ).toString().toLatin1();
            int p = description.indexOf("parameters: ");
            if( p>=0 && description[p+12] )
                params = description.constData() + p + 12;
        }
#endif

#if defined(LZW_SUPPORT) && !defined(NO_LZW_ENCODE) // Oldest (in tiff) most standard lossless scheme.
        uint16 compression = COMPRESSION_LZW;
#elif defined(ZIP_SUPPORT)
        uint16 compression = COMPRESSION_DEFLATE;   // If not compiled, lossless ZIP
        quality = 0xffff;                           // use ZIP's own default.
#elif defined(PACKBITS_SUPPORT)
        uint16 compression = COMPRESSION_PACKBITS;  // Basic RLE (should always be there).
#else
        uint16 compression = COMPRESSION_NONE;      // Else no compression default at all.
#endif

        if( params )
        {
            if( strstr(params, "uncompressed") || strstr(params, "none") )
                compression = COMPRESSION_NONE;

            if( const char* bpsString = strstr(params, "bps=") )
                bps = atoi( bpsString+4 );

#ifdef PACKBITS_SUPPORT
            if( strstr(params, "rle") || strstr(params, "packbits") )
                compression = COMPRESSION_PACKBITS;  // Basic RLE (should always be there).
#endif
#ifdef CCITT_SUPPORT
            if( const char *fax = strstr(params, "fax") )
            {
                if( fax[3] == '=' && fax[4] == '3' )
                    compression = COMPRESSION_CCITTFAX3;
                else
                    compression = COMPRESSION_CCITTFAX4;
            }
#endif
#if defined(LZW_SUPPORT) && !defined(NO_LZW_ENCODE)
            if( strstr(params, "lzw") != 0 )
                compression = COMPRESSION_LZW;
#endif
#ifdef ZIP_SUPPORT
            if( const char *zip = strstr(params, "zip") )
            {
                compression = COMPRESSION_DEFLATE;
                if( zip[3] == '=')
                {
                    quality = atoi(zip+4);
                    if( quality < 1 ) quality = 1;
                    if( quality > 9 ) quality = 9;
                }
                else
                    quality = 0xffff;   // Mark to use libtiff's default.
            }
#endif
#ifdef JPEG_SUPPORT
            if( const char *jpeg = strstr(params, "jpeg") )
            {
                compression = COMPRESSION_JPEG;
                if( jpeg[4] == '=' )
                    quality = atoi(jpeg+5);
            }
#endif
#ifdef PIXARLOG_SUPPORT
            if( const char* pixarlog = strstr(params, "pixarlog") )
            {
                compression = COMPRESSION_PIXARLOG;
                if( pixarlog[8] == '=' )
                    quality = atoi(pixarlog+9);
            }
#endif
#ifdef PIXARFILM_SUPPORT
            if( strstr(params, "pixarfilm") != 0 )
                compression = COMPRESSION_PIXARFILM;
#endif
        }

        if( compression == COMPRESSION_JPEG  && bps != 8 )
        {
            qWarning("Unable to handle write hint bps= with using compression scheme");
            status = -1;
            return;
        }

        int success = TIFFSetField(tif, TIFFTAG_IMAGEWIDTH,      img.width())
                   && TIFFSetField(tif, TIFFTAG_IMAGELENGTH,     img.height())
                   && TIFFSetField(tif, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG)
                   && TIFFSetField(tif, TIFFTAG_ORIENTATION,     ORIENTATION_TOPLEFT)
                   && TIFFSetField(tif, TIFFTAG_COMPRESSION,     compression)
                   ;

        // Add codec specific hints (should be done *after* setting the codec.

        // JPEG quality in 0..100. Use lib's own default if not set.
        if( compression == COMPRESSION_JPEG && quality != 0xffff )
            success &= TIFFSetField(tif, TIFFTAG_JPEGQUALITY, quality);

        // ZIP compression levels 9..1, set from parameters string. The quality
        // is ignored (and variable stolen for compression level), because this
        // is a loosless compression, hence all images have *FULL* quality anyway.
        if( compression == COMPRESSION_DEFLATE && quality != 0xffff )
            success &= TIFFSetField(tif, TIFFTAG_ZIPQUALITY, quality);

        if( img.depth() > 8 )
        {
            bool signedsamples = false;

#if QT_VERSION < 0x040000
            bool alpha   = img.hasAlphaBuffer();
            QImage img32 = img.convertDepth(32);
#else
            bool alpha   = img.hasAlphaChannel();
            QImage img32 = img.convertToFormat( alpha ? QImage::Format_ARGB32 
                                                      : QImage::Format_RGB32 );
#endif

            success &= TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (alpha ? 4 : 3) )
                    && TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,     PHOTOMETRIC_RGB)
                    ;

            // Should setup rowsPerStrip, to have strips of ~ 8KB ~ 64Kb.
            int rowsPerStrip = 8*8192 / (img.width() * bps * (alpha ? 4 : 3));
            rowsPerStrip = (rowsPerStrip/8)*8;    // Make it a multiple of 8 (JPEG compatible).
            if( rowsPerStrip < 8 ) rowsPerStrip = 8;
            if( rowsPerStrip > img.height() ) rowsPerStrip = img.height();

            // For PixarLog 11bit format, make the codec use standard 8bit rgb/rgba input,
            // but still write 11 in the TIFF header...
            if( compression == COMPRESSION_PIXARLOG || compression == COMPRESSION_PIXARFILM )
            {
                // Note: this raise a bug in some versions of the tifflib's
                //       PixarLogVSetField functio, which fails to check
                //       isTiled(tif) before calling TIFFTileSize()...

                if( compression==COMPRESSION_PIXARFILM)
                    success &= TIFFSetField(tif, TIFFTAG_PIXARLOGDATAFMT, PIXARLOGDATAFMT_8BIT);
                else
                    bps = 16;   // Pass linear unsigned 16-bits values to the CODEC.

                // Pass also compression option to the pixarlog compressor.
                if( quality >=1 && quality <=9 )
                    success &= TIFFSetField(tif, TIFFTAG_PIXARLOGQUALITY, quality);
            }
            success &= TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP,  rowsPerStrip)
                    && TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, bps)
                    && TIFFSetField(tif, TIFFTAG_SAMPLEFORMAT,  SAMPLEFORMAT_UINT)
                    ;

            //  Note: TIFFWriteScanline does destroy the buffer passed when a swab
            //        is needed... So always make a copy !
            // The uncompressed buffer needs enough bits to store each components of each pixel,
            // and the compressed one need space rounded only at the end of line level.
            unsigned char *buffer     = new unsigned char[ ((bps+7)/8)*4*img32.width() ],
                          *compressed = new unsigned char[ (bps*4*img32.width() + 7)/8 ];
            Q_CHECK_PTR(buffer);
            Q_CHECK_PTR(compressed);

            for(int y=0; success && (y<img32.height()); ++y)
            {
                unsigned short* words = (unsigned short*)buffer;
                Q_UINT32 *p = (Q_UINT32 *)img32.scanLine(y);
                Q_CHECK_PTR(p);

                if( alpha && bps <= 8)
                {
                    // Make sure to force memory order RGBA, whatever hardware we are
                    // running on:
                    for(int x=0; x<img32.width(); ++x)
                        buffer[4*x  ] = qRed  (p[x]),
                        buffer[4*x+1] = qGreen(p[x]),
                        buffer[4*x+2] = qBlue (p[x]),
                        buffer[4*x+3] = qAlpha(p[x]);
                }
                else if( alpha && bps <= 16 && signedsamples)
                {
                    // We have more that 8 bits. So expand that to signed short first.
                    for(int x=0; x<img32.width(); ++x)
                        words[4*x  ] = ((unsigned int)(qRed  (p[x]) * 0x101))>>1,
                        words[4*x+1] = ((unsigned int)(qGreen(p[x]) * 0x101))>>1,
                        words[4*x+2] = ((unsigned int)(qBlue (p[x]) * 0x101))>>1,
                        words[4*x+3] = ((unsigned int)(qAlpha(p[x]) * 0x101))>>1;
                }
                else if( alpha && bps <= 16)
                {
                    // We have more that 8 bits. So expand that to signed short first.
                    for(int x=0; x<img32.width(); ++x)
                        words[4*x  ] = qRed  (p[x]) * 0x101,
                        words[4*x+1] = qGreen(p[x]) * 0x101,
                        words[4*x+2] = qBlue (p[x]) * 0x101,
                        words[4*x+3] = qAlpha(p[x]) * 0x101;
                }
                else if( bps <= 8)
                {
                    // Make sure to force memory order RGB, whatever hardware we are
                    // running on:
                    for(int x=0; x<img32.width(); ++x)
                        buffer[3*x  ] = qRed  (p[x]),
                        buffer[3*x+1] = qGreen(p[x]),
                        buffer[3*x+2] = qBlue (p[x]);
                }
                else if( bps <= 16 && signedsamples)
                {
                    // We have more that 8 bits. So expand that to signed short first.
                    for(int x=0; x<img32.width(); ++x)
                        words[3*x  ] = ((unsigned int)(qRed  (p[x]) * 0x101))>>1,
                        words[3*x+1] = ((unsigned int)(qGreen(p[x]) * 0x101))>>1,
                        words[3*x+2] = ((unsigned int)(qBlue (p[x]) * 0x101))>>1;
                }
                else if( bps <= 16 )
                {
                    // We have more that 8 bits. So expand that to signed short first.
                    for(int x=0; x<img32.width(); ++x)
                        words[3*x  ] = qRed  (p[x]) * 0x101,
                        words[3*x+1] = qGreen(p[x]) * 0x101,
                        words[3*x+2] = qBlue (p[x]) * 0x101;
                }
                else
                {
                    qWarning("Cannot write more than 16-bit true-color pixels");
                    success=0;
                }

                if( bps < 8 )
                {
                    pack<unsigned char,true>(bps, (alpha ? 4 : 3) * img32.width(), buffer, compressed);
                    success &= ( TIFFWriteScanline(tif, compressed, y) > 0);
                }
                else if( bps > 8 )
                {
                    pack<unsigned short,true>(bps, (alpha ? 4 : 3) * img32.width(), words, compressed);
                    success &= ( TIFFWriteScanline(tif, compressed, y) > 0);
                }
                else
                    success &= ( TIFFWriteScanline(tif, buffer, y) > 0);
            }
            delete[] buffer;
            delete[] compressed;

            status = success ? 0 : -2 ;
        }
        else if( img.depth()==8 )
        {
            success &= TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,   bps)
                    && TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1)
                    ;

            bool useLowBits;

            if( img.isGrayscale() )
            {
                success &= TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
                useLowBits = false;     // This is a color component : truncate lower bits, keep higher ones.
            }
            else
            {
                int colors = img.numColors();

                // In fact, we have no idea of the maximum used color index. So this test might just
                // be the consequence of a 16 color file saved in 8-bit/pixel (aka Photoshop !).
                //if( (1<<bps) < colors )
                //{
                //    qWarning("Cannot write TIFF images with %d colors on %d-bit samples", colors, bps);
                //    success = 0;
                //}

                useLowBits = true;      // This is colors in range 0..(1<<bps - 1) : keep lower bits.

                uint16 *red   = new uint16[colors]; Q_CHECK_PTR(red);
                uint16 *green = new uint16[colors]; Q_CHECK_PTR(green);
                uint16 *blue  = new uint16[colors]; Q_CHECK_PTR(blue);
                float scale=65535.0f/255.0f;
                for(int c=0; c<colors; ++c)
                {
                    QRgb rgb = img.color(c);
                    red[c]   = (int)(qRed(rgb)  *scale + 0.5f);
                    green[c] = (int)(qGreen(rgb)*scale + 0.5f);
                    blue[c]  = (int)(qBlue(rgb) *scale + 0.5f);
                }
                success &= TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
                success &= TIFFSetField(tif, TIFFTAG_COLORMAP,    red, green, blue);
                delete[] blue;
                delete [] green;
                delete[] red;
            }

            // Write all lines, in order (planar does not support anything else ?)

            if( bps == 8 )
            {
                for(int y=0; success && (y<img.height()); ++y)
                    success &= ( TIFFWriteScanline(tif, (tdata_t)img.scanLine(y), y) > 0);
            }
            else if( bps < 8 )
            {
                unsigned char* buffer = new unsigned char[img.width()];
                Q_CHECK_PTR(buffer);

                for(int y=0; y<img.height(); ++y)
                {
                    if( useLowBits )
                        pack<unsigned char,false>(bps,  img.width(), img.scanLine(y), buffer);
                    else
                        pack<unsigned char,true >(bps, img.width(), img.scanLine(y), buffer);
            
                    success &= ( TIFFWriteScanline(tif, (tdata_t)buffer, y, 0) > 0);
                    if( ! success ) break;
                }
                delete[] buffer;
            }
            else if( bps <= 16 )
            {
                unsigned short* buffer     = new unsigned short[img.width()];
                unsigned char * compressed = new unsigned char [(bps*img.width()+7)/8];
                Q_CHECK_PTR(buffer);
                Q_CHECK_PTR(compressed);

                for(int y=0; y<img.height(); ++y)
                {
                    const unsigned char* p = img.scanLine(y);

                    if( useLowBits )
                    {
                        for(int x=0; x<img.width(); ++x)
                            buffer[x] = p[x];
                        pack<unsigned short,false>(bps,  img.width(), buffer, compressed);
                    }
                    else
                    {
                        for(int x=0; x<img.width(); ++x)
                            buffer[x] = p[x]*0x101;
                        pack<unsigned short,true >(bps, img.width(), buffer, compressed);
                    }
            
                    success &= ( TIFFWriteScanline(tif, (tdata_t)compressed, y, 0) > 0);
                    if( ! success ) break;
                }

                delete[] buffer;
                delete[] compressed;
            }

            status = success ? 0 : -2;
        }
        else    // 1-bit bitmap.
        {
            bool msbFirst =
#if QT_VERSION < 0x040000
                (img.bitOrder() == QImage::BigEndian);
#else
                (img.format() == QImage::Format_Mono);
#endif

            success &= TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,   1)
                    && TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1)
                    && TIFFSetField(tif, TIFFTAG_FILLORDER,       msbFirst ? FILLORDER_MSB2LSB : FILLORDER_LSB2MSB)
                    ;

            if( img.color(0) == qRgb(255,255,255) )
                success &= TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
            else
                success &= TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);

            // Write all lines, in order (planar does not support anything else ?)
            for(int y=0; success && (y<img.height()); ++y)
                success &= ( TIFFWriteScanline(tif, (tdata_t)img.scanLine(y), y) > 0);

            status = success ? 0 : -2;
        }

        TIFFFlush(tif);
        TIFFClose(tif);

    }
    else
        status = -1;
}
